# Copyright 2014 Google Inc. All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""User-facing interface to an ADB device.

This module provides the user-facing interface to controlling ADB devices with
this library.  See adb_protocol.py for a breakdown of the various layers.
Essentially, this module provides the AdbDevice class, which takes a UsbHandle
and uses the layers in between to provide a high level interface that is
designed to be similar to the 'adb' commandline interface.

Some additional useful mechanisms are provided for more common programmatic
needs, see shell_service.py for a description of how to use asynchronous shell
commands.

Note we don't provide a device listing facility, as this is not actually an ADB
function, but rather a USB function - listing devices with a specific interface
class, subclass, and protocol.
"""

import logging
import os.path

try:
  from M2Crypto import RSA
except ImportError:
  logging.error('Failed to import M2Crypto, did you pip install '
                'openhtf[usb_plugs]?')
  raise

from openhtf.plugs.usb import adb_protocol
from openhtf.plugs.usb import filesync_service
from openhtf.plugs.usb import shell_service
from openhtf.plugs.usb import usb_exceptions

from openhtf.util import timeouts
import six

# USB interface class, subclass, and protocol for matching against.
CLASS = 0xFF
SUBCLASS = 0x42
PROTOCOL = 0x01


class M2CryptoSigner(adb_protocol.AuthSigner):
  """AuthSigner using M2Crypto."""

  def __init__(self, rsa_key_path):
    with open(rsa_key_path + '.pub') as rsa_pub_file:
      self.public_key = rsa_pub_file.read()

    self.rsa_key = RSA.load_key(rsa_key_path)

  def sign(self, data):
    return self.rsa_key.sign(data, 'sha1')

  def get_public_key(self):
    """Return the public key."""
    return self.public_key


class AdbDevice(object):
  """Exposes adb-like methods for use.

  Some methods are more-pythonic and/or have more options.

  Attributes:
    filesync_service: Direct access to this device's FilesyncService.
    shell_service: Direct access to this device's ShellService.
  """

  def __init__(self, adb_connection):
    self._adb_connection = adb_connection
    self.filesync_service = filesync_service.FilesyncService.using_connection(
        adb_connection)
    self.shell_service = shell_service.ShellService.using_connection(
        adb_connection)

  def __str__(self):
    return '<%s: %s(%s) @%s>' % (type(self).__name__,
                                 self._adb_connection.serial,
                                 self._adb_connection.systemtype,
                                 self._adb_connection.transport)
  __repr__ = __str__

  def get_system_type(self):
    """Return the system type."""
    return self._adb_connection.systemtype

  def get_serial(self):
    """Return the device serial."""
    return self._adb_connection.serial

  def close(self):
    """Close the ADB connection."""
    self.filesync_service.close()
    self._adb_connection.close()

  def install(self, apk_path, destination_dir=None, timeout_ms=None):
    """Install apk to device.

    Doesn't support verifier file, instead allows destination directory to be
    overridden.

    Arguments:
      apk_path: Local path to apk to install.
      destination_dir: Optional destination directory. Use /system/app/ for
        persistent applications.
      timeout_ms: Expected timeout for pushing and installing.

    Returns:
      The pm install output.
    """
    if not destination_dir:
      destination_dir = '/data/local/tmp/'
    basename = os.path.basename(apk_path)
    destination_path = destination_dir + basename
    self.push(apk_path, destination_path, timeout_ms=timeout_ms)
    return self.Shell('pm install -r "%s"' % destination_path,
                      timeout_ms=timeout_ms)

  def push(self, source_file, device_filename, timeout_ms=None):
    """Push source_file to file on device.

    Arguments:
      source_file: Either a filename or file-like object to push to the device.
        If a filename, will set the remote mtime to match the local mtime,
        otherwise will use the current time.
      device_filename: The filename on the device to write to.
      timeout_ms: Expected timeout for any part of the push.
    """
    mtime = 0
    if isinstance(source_file, six.string_types):
      mtime = os.path.getmtime(source_file)
      source_file = open(source_file)

    self.filesync_service.send(
        source_file, device_filename, mtime=mtime,
        timeout=timeouts.PolledTimeout.from_millis(timeout_ms))

  def pull(self, device_filename, dest_file=None, timeout_ms=None):
    """Pull file from device.

    Arguments:
      device_filename: The filename on the device to pull.
      dest_file: If set, a filename or writable file-like object.
      timeout_ms: Expected timeout for the pull.

    Returns:
      The file data if dest_file is not set, None otherwise.
    """
    should_return_data = dest_file is None
    if isinstance(dest_file, six.string_types):
      dest_file = open(dest_file, 'w')
    elif dest_file is None:
      dest_file = six.StringIO()
    self.filesync_service.recv(device_filename, dest_file,
                               timeouts.PolledTimeout.from_millis(timeout_ms))
    if should_return_data:
      return dest_file.getvalue()

  def list(self, device_path, timeout_ms=None):
    """Yield filesync_service.DeviceFileStat objects for directory contents."""
    return self.filesync_service.list(
        device_path, timeouts.PolledTimeout.from_millis(timeout_ms))

  def command(self, command, raw=False, timeout_ms=None):
    """Run command on the device, returning the output."""
    return self.shell_service.command(
        str(command), raw=raw, timeout_ms=timeout_ms)

  Shell = command  #pylint: disable=invalid-name

  def async_command(self, command, raw=False, timeout_ms=None):
    """See shell_service.ShellService.async_command()."""
    return self.shell_service.async_command(
        str(command), raw=raw, timeout_ms=timeout_ms)

  def _check_remote_command(self, destination, timeout_ms, success_msgs=None):
    """Open a stream to destination, check for remote errors.

    Used for reboot, remount, and root services.  If this method returns, the
    command was successful, otherwise an appropriate error will have been
    raised.

    Args:
      destination: Stream destination to open.
      timeout_ms: Timeout in milliseconds for the operation.
      success_msgs: If provided, a list of messages that, if returned from the
        device, indicate success, so don't treat them as errors.

    Raises:
      AdbRemoteError: If the remote command fails, will contain any message we
        got back from the device.
      AdbStreamUnavailableError: The service requested isn't supported.
    """
    timeout = timeouts.PolledTimeout.from_millis(timeout_ms)
    stream = self._adb_connection.open_stream(destination, timeout)
    if not stream:
      raise usb_exceptions.AdbStreamUnavailableError(
          'Service %s not supported', destination)
    try:
      message = stream.read(timeout_ms=timeout)
      # Some commands report success messages, ignore them.
      if any([m in message for m in success_msgs]):
        return
    except usb_exceptions.CommonUsbError:
      if destination.startswith('reboot:'):
        # We expect this if the device is rebooting.
        return
      raise
    raise usb_exceptions.AdbRemoteError('Device message: %s', message)

  def reboot(self, destination='', timeout_ms=None):
    """Reboot device, specify 'bootloader' for fastboot."""
    self._check_remote_command('reboot:%s' % destination, timeout_ms)

  def remount(self, timeout_ms=None):
    """Remount / as read-write."""
    self._check_remote_command('remount:', timeout_ms, ['remount succeeded'])

  def root(self, timeout_ms=None):
    """Restart adbd as root on device."""
    self._check_remote_command('root:', timeout_ms,
                               ['already running as root',
                                'restarting adbd as root'])

  @classmethod
  def connect(cls, usb_handle, **kwargs):
    """Connect to the device.

    Args:
      usb_handle: UsbHandle instance to use.
      **kwargs: See AdbConnection.connect for kwargs. Includes rsa_keys, and
        auth_timeout_ms.

    Returns:
      An instance of this class if the device connected successfully.
    """
    adb_connection = adb_protocol.AdbConnection.connect(usb_handle, **kwargs)
    return cls(adb_connection)



import random

def auto_graph_to_plotly_fig(graph, seed=98):
    import random
    random.seed(seed) # Always the same seed to generate the same graph everytime
    
    coord_data = {}
    coords = {}
    
    def coord_fn(data):
        qualname = data['qualname']
        if qualname not in coords:
            depth = len(qualname.split('.'))
            x = coord_data.get(depth, 0)
            y = depth + random.uniform(0, 0.50)*(x%2)
            
            x = x - random.uniform(0, 0.50)*((y-1)%2)
            
            coord_data[depth] = coord_data.get(depth, 0) + 1
            coords[qualname] = (x, y)
        return coords[qualname]
    
    return graph_to_plotly_fig(graph, coord_fn)

def graph_to_plotly_fig(graph, coord_fn, text_fn=None):
    import plotly.graph_objects as go
    
    
    annotations = []
    
    edge_x = []
    edge_y = []
    edge_text = []
    
    for edge in graph.edges():
        
        x0, y0 = coord_fn(graph.node[edge[0]])
        x1, y1 = coord_fn(graph.node[edge[1]])
        annotations.append(go.layout.Annotation(
            x=x1,
            y=y1,
            showarrow=True,
            arrowhead=2,
            arrowsize=4,
            xref="x",
            yref="y",
            axref="x",
            ayref="y",
            ax=x0,
            ay=y0,
        ))
        
    
    # edge_trace = go.Scatter(
    #     x=edge_x, y=edge_y,
    #     line=dict(width=0.5, color='#888'),
    #     hoverinfo='none',
    #     mode='lines')

    node_x = []
    node_y = []
    colors = []
    # node_text = []
    for nodename in graph.nodes():
        node = graph.node[nodename]
        x, y = coord_fn(node)
        node_x.append(x)
        node_y.append(y)
        if text_fn:
            text = text_fn(node)
        else:
            text = nodename
        
        if node['node'].testable:
            color = 'green'
        else:
            color = 'grey'
        colors.append(color)
        # node_text.append(text)
        annotations.append(go.layout.Annotation(
            x=x,
            y=y,
            xref="x",
            yref="y",
            text=text,
            ax=0,
            ay=-40
        ))

    node_trace = go.Scatter(
        x=node_x, y=node_y,
        mode='markers',
        hoverinfo='text',
        marker=dict(
            showscale=True,
            # colorscale options
            #'Greys' | 'YlGnBu' | 'Greens' | 'YlOrRd' | 'Bluered' | 'RdBu' |
            #'Reds' | 'Blues' | 'Picnic' | 'Rainbow' | 'Portland' | 'Jet' |
            #'Hot' | 'Blackbody' | 'Earth' | 'Electric' | 'Viridis' |
            colorscale='YlGnBu',
            reversescale=True,
            color=colors,
            size=10,
            colorbar=dict(
                thickness=15,
                title='Node Connections',
                xanchor='left',
                titleside='right'
            ),
            line_width=2))
    
    # node_trace.text = node_text
    
    fig = go.Figure(data=[node_trace],
             layout=go.Layout(
                title='Network graph made with Python',
                titlefont_size=16,
                showlegend=False,
                hovermode='closest',
                margin=dict(b=20,l=5,r=5,t=40),
                annotations=annotations,
                xaxis=dict(showgrid=False, zeroline=False, showticklabels=False),
                yaxis=dict(showgrid=False, zeroline=False, showticklabels=False))
                )

    return fig


# Copyright 2014 Google Inc. All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Common exceptions for USB, ADB and Fastboot."""

import logging

try:
  import libusb1
except ImportError:
  logging.error('Failed to import libusb, did you pip install '
                'openhtf[usb_plugs]?')
  raise

_LOG = logging.getLogger(__name__)


class CommonUsbError(Exception):
  """Exception that both looks good and is functional.

  Okay, not that kind of functional, it's still a class.

  This interpolates the message with the given arguments to make it
  human-readable, but keeps the arguments in case other code try-excepts it.
  """

  def __init__(self, message=None, *args):
    if message is not None:
      if '%' in message:
        try:
          message %= args
        except TypeError:
          # This is a fairly obscure failure, so we intercept it and emit a
          # more useful error message.
          _LOG.error('USB Exceptions expect a format-string, do not include '
                     'percent symbols to disable this functionality: %s',
                     message)
          raise
      super(CommonUsbError, self).__init__(message, *args)
    else:
      super(CommonUsbError, self).__init__(*args)


# USB exceptions, these are not specific to any particular protocol.
class LibusbWrappingError(CommonUsbError):
  """Wraps libusb1 errors while keeping its original usefulness.

  Attributes:
    usb_error: Instance of libusb1.USBError
  """

  def __init__(self, usb_error, *args):
    super(LibusbWrappingError, self).__init__(*args)
    self.usb_error = usb_error

  def __str__(self):
    return '<%s: %s>' % (type(self).__name__, str(self.usb_error))

  def is_timeout(self):
    """Returns True if the USBError we are wrapping is a timeout error."""
    return self.usb_error.value == libusb1.LIBUSB_ERROR_TIMEOUT


class InvalidArgumentError(CommonUsbError):
  """Raised when an invalid argument is passed to a USB function."""


class DeviceNotFoundError(CommonUsbError):
  """Device isn't on USB."""


class HandleClosedError(CommonUsbError):
  """Raised when an operation is requested on an already closed handle."""


class InvalidEndpointsError(CommonUsbError):
  """Raised when a USB interface does not have supported endpoints."""


class MultipleInterfacesFoundError(CommonUsbError):
  """Raised when multiple matching devices are found for an Open call."""


class UsbServiceCommunicationError(CommonUsbError):
  """Could not connect to usb service."""


class UnclosedHandleError(CommonUsbError):
  """Raised when a UsbHandle goes out of scope but hasn't been Close()'d."""


class UsbWriteFailedError(LibusbWrappingError):
  """Raised when a USB write fails."""


class UsbReadFailedError(LibusbWrappingError):
  """Raised when a USB read fails."""


# ADB specific exceptions
class DeviceAuthError(CommonUsbError):
  """Device authentication failed."""


class AdbOperationException(Exception):
  """Failed to communicate over adb with device after multiple retries."""


class AdbDataIntegrityError(CommonUsbError):
  """Raised when an AdbMessage has corrupted data."""


class AdbProtocolError(CommonUsbError):
  """Raised when we don't get the type of ADB packet we expect."""


class AdbStreamClosedError(CommonUsbError):
  """Raised when an attempt to read/write happens on a closed AdbStream."""


class AdbStreamUnavailableError(CommonUsbError):
  """Raised when a stream cannot be opened."""


class AdbTransportClosedError(CommonUsbError):
  """Raised when a transport (a layer above a Stream) is closed."""


class AdbRemoteError(CommonUsbError):
  """Raised when there's an error on the remote (device) side."""


class AdbTimeoutError(CommonUsbError):
  """Raised when an ADB operation times out."""


# Fastboot specific exceptions
class FastbootTransferError(CommonUsbError):
  """Transfer error."""


class FastbootRemoteFailure(CommonUsbError):
  """Remote error."""


class FastbootStateMismatch(CommonUsbError):
  """Fastboot and uboot's state machines are arguing. You Lose."""


class FastbootInvalidResponse(CommonUsbError):
  """Fastboot responded with a header we didn't expect."""



import openhtf.plugs as plugs
from openhtf.util import conf

import time

import serial
from serial.threaded import ReaderThread

from .base import IOInterface

class ComportInterface(IOInterface):
    """An interface to a comport.
    
    Allows reading and writing. A background thread reads any data that comes in 
    and those lines can be accessed using the `next_line` function.
    
    """

    def __init__(self, comport, baudrate=115200):
        super().__init__()
        self.comport = comport
        self.baudrate = baudrate

        self._serial = None
        self._reader = None

    def open(self, _serial=None):
        """Opens the serial port using the :attr:`~ComportInterface.comport` and :attr:`~ComportInterface.baudrate`
        object attributes.
        
        Arguments:
            _serial:
                Optionnal underlying :class:`serial.Serial` object to use. Used for mock testing.
        
        """
        self.close()
        if not _serial:
            _serial = serial.Serial(self.comport, self.baudrate, timeout=self.timeout)
        self._serial = _serial
        self._reader = ReaderThread(self._serial, lambda: self)
        self._reader.start()

    def close(self):
        """Attempts to close the serial port if it exists."""
        if self._reader is not None and self._serial.is_open:
            self._reader.close()
            self._reader = None

    def _write(self, string):
        """Write the string into the comport."""
        return self._reader.write(string.encode('utf8'))

    def com_target(self, *args, **kwargs):
        """Alias for message_target"""
        return self.message_target(*args, **kwargs)



import sys
import argparse

import openhtf as htf

from openhtf.core.test_descriptor import create_arg_parser

def spintop_create_arg_parser(add_help=False):
    parser = argparse.ArgumentParser(
          'Spintop OpenHTF - The extension to OpenHTF', parents=[create_arg_parser()],
          add_help=add_help)
    parser.add_argument(
        '--coverage', action='store_true',
        help='Print the list of covered components')
    return parser

class Test(htf.Test):
    ARG_PARSER_FACTORY = spintop_create_arg_parser

    def configure(self, **kwargs):
        # This calls known_args_hook
        super(Test, self).configure(**kwargs)

    def known_args_hook(self, known_args):
        """ Check for known args and perform possible effect instead of executing the test.
        Cancel execution using sys.exit(0)
        """
        if known_args.coverage:
            sys.exit(0)
        super(Test, self).known_args_hook(known_args)



# Copyright 2016 Google Inc. All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Example OpenHTF stop_on_first_failure test executor option.

This is feature is very useful when you do not want the test to continue
execution when the test result is expected to be a FAIL due to a failed
phase result. This feature can also be very useful during debug sessions.

Two options are provided to set this option.
1. Using test.configure
test.configure(stop_on_first_failure=True)

2. Using config item. This option lets you toggle this feature dynamically.
conf.load(stop_on_first_failure=True)
"""


import openhtf as htf

from openhtf.plugs import user_input
from openhtf.util import conf  # pylint: disable=unused-import
from openhtf.util import validators
from openhtf.output.callbacks import console_summary


@htf.measures('number_sum', validators=[validators.in_range(0, 5)])
def add_numbers_fails(test):
  """Add numbers fails phase

  This phase will return a failed measurement number_sum.
  """
  test.logger.info('Add numbers 2 and 4')
  number_sum = 2 + 4
  test.measurements.number_sum = number_sum


# This is phase is not expected to run as the add_numbers phase is
# expected to fail.
@htf.measures(htf.Measurement('hello_world_measurement'))
def hello_world(test):
  """A hello world test phase."""

  test.logger.info('This phase will not be run since previous phase failed')
  test.measurements.hello_world_measurement = 'Hello World!'


if __name__ == '__main__':
  test = htf.Test(add_numbers_fails, hello_world)
  test.add_output_callbacks(console_summary.ConsoleSummary())
  # Option 1: test.configure
  test.configure(stop_on_first_failure=True)

  # Option 2 : You can disable option 1 and enable below line
  # to get same result
  # conf.load(stop_on_first_failure=True)

  test.execute(test_start=user_input.prompt_for_test_start())


# Copyright 2018 Google Inc. All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Console output utilities for OpenHTF.

This module provides convenience methods to format output for the CLI, along
with the ability to fork output to both a logger (e.g. a test record logger)
and to the CLI directly (e.g. sys.stdout).

Under the default configuration, messages printed with these utilities will be
saved to the test record logs of all running tests. To change this behavior,
the `logger` parameter should be overridden, e.g. by passing in the test record
logger for the current test.
"""

import logging
import math
import os
import re
import string
import sys
import textwrap
import time

import colorama
import contextlib2 as contextlib

from openhtf.util import argv

# Colorama module has to be initialized before use.
colorama.init()

_LOG = logging.getLogger(__name__)

# If True, all CLI output through this module will be suppressed, as well as any
# logging that uses a CliQuietFilter.
CLI_QUIET = False

ARG_PARSER = argv.ModuleParser()
ARG_PARSER.add_argument(
    '--quiet', action=argv.StoreTrueInModule, target='%s.CLI_QUIET' % __name__,
    help=textwrap.dedent('''\
        Suppress all CLI output from OpenHTF's printing functions and logging.
        This flag will override any verbosity levels set with -v.'''))

ANSI_ESC_RE = re.compile(r'\x1B\[[0-?]*[ -/]*[@-~]')


class ActionFailedError(Exception):
  """Indicates an action failed. Used internally by action_result_context()."""


def _printed_len(some_string):
  """Compute the visible length of the string when printed."""
  return len([x for  x in ANSI_ESC_RE.sub('', some_string)
              if x in string.printable])


def _linesep_for_file(file):
  """Determine which line separator to use based on the file's mode."""
  if 'b' in file.mode:
    return os.linesep
  return '\n'


def banner_print(msg, color='', width=60, file=sys.stdout, logger=_LOG):
  """Print the message as a banner with a fixed width.

  Also logs the message (un-bannered) to the given logger at the debug level.

  Args:
    msg: The message to print.
    color: Optional colorama color string to be applied to the message. You can
        concatenate colorama color strings together in order to get any set of
        effects you want.
    width: Total width for the resulting banner.
    file: A file object to which the banner text will be written. Intended for
        use with CLI output file objects like sys.stdout.
    logger: A logger to use, or None to disable logging.

  Example:

    >>> banner_print('Foo Bar Baz')

    ======================== Foo Bar Baz =======================

  """
  if logger:
    logger.debug(ANSI_ESC_RE.sub('', msg))
  if CLI_QUIET:
    return
  lpad = int(math.ceil((width - _printed_len(msg) - 2) / 2.0)) * '='
  rpad = int(math.floor((width - _printed_len(msg) - 2) / 2.0)) * '='
  file.write('{sep}{color}{lpad} {msg} {rpad}{reset}{sep}{sep}'.format(
      sep=_linesep_for_file(file), color=color, lpad=lpad, msg=msg, rpad=rpad,
      reset=colorama.Style.RESET_ALL))
  file.flush()


def bracket_print(msg, color='', width=8, file=sys.stdout):
  """Prints the message in brackets in the specified color and end the line.

  Args:
    msg: The message to put inside the brackets (a brief status message).
    color: Optional colorama color string to be applied to the message. You can
        concatenate colorama color strings together in order to get any set of
        effects you want.
    width: Total desired width of the bracketed message.
    file: A file object to which the bracketed text will be written. Intended
        for use with CLI output file objects like sys.stdout.
    """
  if CLI_QUIET:
    return
  lpad = int(math.ceil((width - 2 - _printed_len(msg)) / 2.0)) * ' '
  rpad = int(math.floor((width - 2 - _printed_len(msg)) / 2.0)) * ' '
  file.write('[{lpad}{bright}{color}{msg}{reset}{rpad}]'.format(
      lpad=lpad, bright=colorama.Style.BRIGHT, color=color, msg=msg,
      reset=colorama.Style.RESET_ALL, rpad=rpad))
  file.write(colorama.Style.RESET_ALL)
  file.write(_linesep_for_file(file))
  file.flush()


def cli_print(msg, color='', end=None, file=sys.stdout, logger=_LOG):
  """Print the message to file and also log it.

  This function is intended as a 'tee' mechanism to enable the CLI interface as
  a first-class citizen, while ensuring that everything the operator sees also
  has an analogous logging entry in the test record for later inspection.

  Args:
    msg: The message to print/log.
    color: Optional colorama color string to be applied to the message. You can
        concatenate colorama color strings together in order to get any set of
        effects you want.
    end: A custom line-ending string to print instead of newline.
    file: A file object to which the baracketed text will be written. Intended
        for use with CLI output file objects like sys.stdout.
    logger: A logger to use, or None to disable logging.
  """
  if logger:
    logger.debug('-> {}'.format(msg))
  if CLI_QUIET:
    return
  if end is None:
    end = _linesep_for_file(file)
  file.write('{color}{msg}{reset}{end}'.format(
      color=color, msg=msg, reset=colorama.Style.RESET_ALL, end=end))


def error_print(msg, color=colorama.Fore.RED, file=sys.stderr):
  """Print the error message to the file in the specified color.

  Args:
    msg: The error message to be printed.
    color: Optional colorama color string to be applied to the message. You can
        concatenate colorama color strings together here, but note that style
        strings will not be applied.
    file: A file object to which the baracketed text will be written. Intended
        for use with CLI output file objects, specifically sys.stderr.
  """
  if CLI_QUIET:
    return
  file.write('{sep}{bright}{color}Error: {normal}{msg}{sep}{reset}'.format(
      sep=_linesep_for_file(file), bright=colorama.Style.BRIGHT, color=color,
      normal=colorama.Style.NORMAL, msg=msg, reset=colorama.Style.RESET_ALL))
  file.flush()


class ActionResult(object):
  """Used with an action_result_context to signal the result of an action."""

  def __init__(self):
    self.success = None

  def succeed(self):
    """Mark the action as having succeeded."""
    self.success = True

  def fail(self):
    """Mark the action as having failed.

    Raises:
      ActionFailedError: This exception is always raised, by this function,
          but should be caught by the contextmanager.
    """
    self.success = False
    raise ActionFailedError()


@contextlib.contextmanager
def action_result_context(action_text,
                          width=60,
                          status_width=8,
                          succeed_text='OK',
                          fail_text='FAIL',
                          unknown_text='????',
                          file=sys.stdout,
                          logger=_LOG):
  """A contextmanager that prints actions and results to the CLI.

  When entering the context, the action will be printed, and when the context
  is exited, the result will be printed. The object yielded by the context is
  used to mark the action as a success or failure, and a raise from inside the
  context will also result in the action being marked fail. If the result is
  left unset, then indicative text ("????") will be printed as the result.

  Args:
    action_text: Text to be displayed that describes the action being taken.
    width: Total width for each line of output.
    status_width: Width of the just the status message portion of each line.
    succeed_text: Status message displayed when the action succeeds.
    fail_text: Status message displayed when the action fails.
    unknown_text: Status message displayed when the result is left unset.
    file: Specific file object to write to write CLI output to.
    logger: A logger to use, or None to disable logging.

  Example usage:
    with action_result_context('Doing an action that will succeed...') as act:
      time.sleep(2)
      act.succeed()

    with action_result_context('Doing an action with unset result...') as act:
      time.sleep(2)

    with action_result_context('Doing an action that will fail...') as act:
      time.sleep(2)
      act.fail()

    with action_result_context('Doing an action that will raise...') as act:
      time.sleep(2)
      import textwrap
      raise RuntimeError(textwrap.dedent('''\
          Uh oh, looks like there was a raise in the mix.

          If you see this message, it means you are running the console_output
          module directly rather than using it as a library. Things to try:

            * Not running it as a module.
            * Running it as a module and enjoying the preview text.
            * Getting another coffee.'''))

  Example output:
    Doing an action that will succeed...                [  OK  ]
    Doing an action with unset result...                [ ???? ]
    Doing an action that will fail...                   [ FAIL ]
    Doing an action that will raise...                  [ FAIL ]
    ...
  """
  if logger:
    logger.debug('Action - %s', action_text)
  if not CLI_QUIET:
    file.write(''.join((action_text, '\r')))
    file.flush()
    spacing = (width - status_width - _printed_len(action_text)) * ' '

  result = ActionResult()
  try:
    yield result
  except Exception as err:
    if logger:
      logger.debug('Result - %s [ %s ]', action_text, fail_text)
    if not CLI_QUIET:
      file.write(''.join((action_text, spacing)))
      bracket_print(fail_text, width=status_width, color=colorama.Fore.RED,
                    file=file)
    if not isinstance(err, ActionFailedError):
      raise
    return

  result_text = succeed_text if result.success else unknown_text
  result_color = colorama.Fore.GREEN if result.success else colorama.Fore.YELLOW
  if logger:
    logger.debug('Result - %s [ %s ]', action_text, result_text)
  if not CLI_QUIET:
    file.write(''.join((action_text, spacing)))
    bracket_print(result_text, width=status_width, color=result_color,
                  file=file)


# If invoked as a runnable module, this module will invoke its action result
# context in order to print colorized example output.
if __name__ == '__main__':
  banner_print('Running pre-flight checks.')

  with action_result_context('Doing an action that will succeed...') as act:
    time.sleep(2)
    act.succeed()

  with action_result_context('Doing an action with unset result...') as act:
    time.sleep(2)

  with action_result_context('Doing an action that will fail...') as act:
    time.sleep(2)
    act.fail()

  with action_result_context('Doing an action that will raise...') as act:
    time.sleep(2)
    raise RuntimeError(textwrap.dedent('''\
        Uh oh, looks like there was a raise in the mix.

        If you see this message, it means you are running the console_output
        module directly rather than using it as a library. Things to try:

          * Not running it as a module.
          * Running it as a module and enjoying the preview text.
          * Getting another coffee.'''))


class CliQuietFilter(logging.Filter):
  """Filter that suppresses logging output if the --quiet CLI option is set.

  Note that the --quiet CLI option is stored in the CLI_QUIET member of this
  module, and can thus be overridden in test scripts. This filter should only
  be used with loggers that print to the CLI.
  """
  def filter(self, record):
    return not CLI_QUIET



# Copyright 2014 Google Inc. All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""A libusb1-based fastboot implementation."""

import binascii
import collections
import logging
import os
import struct

from . import usb_exceptions
from openhtf.util import argv
import six


FASTBOOT_DOWNLOAD_CHUNK_SIZE_KB = 1024

ARG_PARSER = argv.ModuleParser()
ARG_PARSER.add_argument(
    '--fastboot_download_chunk_size_kb',
    default=FASTBOOT_DOWNLOAD_CHUNK_SIZE_KB,
    action=argv.StoreInModule,
    type=int,
    target='%s.FASTBOOT_DOWNLOAD_CHUNK_SIZE_KB' % __name__,
    help='Size of chunks to send when downloading fastboot images')

_LOG = logging.getLogger(__name__)

DEFAULT_MESSAGE_CALLBACK = lambda m: _LOG.info('Got %s from device', m)
FastbootMessage = collections.namedtuple(  # pylint: disable=invalid-name
    'FastbootMessage', ['message', 'header'])


class FastbootProtocol(object):
  """Encapsulates the fastboot protocol."""
  FINAL_HEADERS = {'OKAY', 'DATA'}

  def __init__(self, usb):
    """Constructs a FastbootProtocol instance.

    Arguments:
      usb: UsbHandle instance.
    """
    self.usb = usb

  @property
  def usb_handle(self):
    """This instance's USB handle."""
    return self.usb

  def send_command(self, command, arg=None):
    """Sends a command to the device.

    Args:
      command: The command to send.
      arg: Optional argument to the command.
    """
    if arg is not None:
      command = '%s:%s' % (command, arg)
    self._write(six.StringIO(command), len(command))

  def handle_simple_responses(
      self, timeout_ms=None, info_cb=DEFAULT_MESSAGE_CALLBACK):
    """Accepts normal responses from the device.

    Args:
      timeout_ms: Timeout in milliseconds to wait for each response.
      info_cb: Optional callback for text sent from the bootloader.

    Returns:
      OKAY packet's message.
    """
    return self._accept_responses('OKAY', info_cb, timeout_ms=timeout_ms)

  # pylint: disable=too-many-arguments
  def handle_data_sending(self, source_file, source_len,
                          info_cb=DEFAULT_MESSAGE_CALLBACK,
                          progress_callback=None, timeout_ms=None):
    """Handles the protocol for sending data to the device.

    Arguments:
      source_file: File-object to read from for the device.
      source_len: Amount of data, in bytes, to send to the device.
      info_cb: Optional callback for text sent from the bootloader.
      progress_callback: Callback that takes the current and the total progress
        of the current file.
      timeout_ms: Timeout in milliseconds to wait for each response.

    Raises:
      FastbootTransferError: When fastboot can't handle this amount of data.
      FastbootStateMismatch: Fastboot responded with the wrong packet type.
      FastbootRemoteFailure: Fastboot reported failure.
      FastbootInvalidResponse: Fastboot responded with an unknown packet type.

    Returns:
      OKAY packet's message.
    """
    accepted_size = self._accept_responses(
        'DATA', info_cb, timeout_ms=timeout_ms)

    accepted_size = binascii.unhexlify(accepted_size[:8])
    accepted_size, = struct.unpack('>I', accepted_size)
    if accepted_size != source_len:
      raise usb_exceptions.FastbootTransferError(
          'Device refused to download %s bytes of data (accepts %s bytes)',
          source_len, accepted_size)
    self._write(source_file, accepted_size, progress_callback)
    return self._accept_responses('OKAY', info_cb, timeout_ms=timeout_ms)

  # pylint: enable=too-many-arguments

  def _accept_responses(self, expected_header, info_cb, timeout_ms=None):
    """Accepts responses until the expected header or a FAIL.

    Arguments:
      expected_header: OKAY or DATA
      info_cb: Optional callback for text sent from the bootloader.
      timeout_ms: Timeout in milliseconds to wait for each response.

    Raises:
      FastbootStateMismatch: Fastboot responded with the wrong packet type.
      FastbootRemoteFailure: Fastboot reported failure.
      FastbootInvalidResponse: Fastboot responded with an unknown packet type.

    Returns:
      OKAY packet's message.
    """
    while True:
      response = self.usb.read(64, timeout_ms=timeout_ms)
      header = response[:4]
      remaining = response[4:]

      if header == 'INFO':
        info_cb(FastbootMessage(remaining, header))
      elif header in self.FINAL_HEADERS:
        if header != expected_header:
          raise usb_exceptions.FastbootStateMismatch(
              'Expected %s, got %s', expected_header, header)
        if header == 'OKAY':
          info_cb(FastbootMessage(remaining, header))
        return remaining
      elif header == 'FAIL':
        info_cb(FastbootMessage(remaining, header))
        raise usb_exceptions.FastbootRemoteFailure('FAIL: %s', remaining)
      else:
        raise usb_exceptions.FastbootInvalidResponse(
            'Got unknown header %s and response %s', header, remaining)

  def _handle_progress(self, total, progress_callback):  # pylint: disable=no-self-use
    """Calls the callback with the current progress and total ."""
    current = 0
    while True:
      current += yield
      try:
        progress_callback(current, total)
      except Exception:  # pylint: disable=broad-except
        _LOG.exception('Progress callback raised an exception. %s',
                       progress_callback)
        continue

  def _write(self, data, length, progress_callback=None):
    """Sends the data to the device, tracking progress with the callback."""
    if progress_callback:
      progress = self._handle_progress(length, progress_callback)
      six.next(progress)
    while length:
      tmp = data.read(FASTBOOT_DOWNLOAD_CHUNK_SIZE_KB * 1024)
      length -= len(tmp)
      self.usb.write(tmp)

      if progress_callback:
        progress.send(len(tmp))


class FastbootCommands(object):
  """Encapsulates the fastboot commands."""
  protocol_handler = FastbootProtocol

  def __init__(self, usb):
    """Constructs a FastbootCommands instance.

    Arguments:
      usb: UsbHandle instance.
    """
    self._usb = usb
    self._protocol = self.protocol_handler(usb)

  @property
  def handle(self):
    """This instance's USB handle."""
    return self._usb

  def close(self):
    """Close the USB handle."""
    self._usb.close()

  def _simple_command(self, command, arg=None, **kwargs):
    """Send a simple command."""
    self._protocol.send_command(command, arg)
    return self._protocol.handle_simple_responses(**kwargs)

  # pylint: disable=too-many-arguments
  def flash_from_file(self, partition, source_file, source_len=0,
                      info_cb=DEFAULT_MESSAGE_CALLBACK, progress_callback=None,
                      timeout_ms=None):
    """Flashes a partition from the file on disk.

    Args:
      partition: Partition name to flash to.
      source_file: Filename to download to the device.
      source_len: Optional length of source_file, uses os.stat if not provided.
      info_cb: See Download.
      progress_callback: See Download.
      timeout_ms: The amount of time to wait on okay after flashing.

    Returns:
      Download and flash responses, normally nothing.
    """
    if source_len == 0:
      # Fall back to stat.
      source_len = os.stat(source_file).st_size
    download_response = self.download(
        source_file, source_len=source_len, info_cb=info_cb,
        progress_callback=progress_callback)
    flash_response = self.flash(partition, info_cb=info_cb,
                                timeout_ms=timeout_ms)
    return download_response + flash_response

  # pylint: enable=too-many-arguments

  def download(self, source_file, source_len=0,
               info_cb=DEFAULT_MESSAGE_CALLBACK, progress_callback=None):
    """Downloads a file to the device.

    Args:
      source_file: A filename or file-like object to download to the device.
      source_len: Optional length of source_file. If source_file is a file-like
          object and source_len is not provided, source_file is read into
          memory.
      info_cb: Optional callback accepting FastbootMessage for text sent from
          the bootloader.
      progress_callback: Optional callback called with the percent of the
          source_file downloaded. Note, this doesn't include progress of the
          actual flashing.

    Returns:
      Response to a download request, normally nothing.
    """
    if isinstance(source_file, six.string_types):
      source_len = os.stat(source_file).st_size
      source_file = open(source_file)

    if source_len == 0:
      # Fall back to storing it all in memory :(
      data = source_file.read()
      source_file = six.StringIO(data)
      source_len = len(data)

    self._protocol.send_command('download', '%08x' % source_len)
    return self._protocol.handle_data_sending(
        source_file, source_len, info_cb, progress_callback=progress_callback)

  def flash(self, partition, timeout_ms=None, info_cb=DEFAULT_MESSAGE_CALLBACK):
    """Flashes the last downloaded file to the given partition.

    Args:
      partition: Partition to flash.
      timeout_ms: Optional timeout in milliseconds to wait for it to finish.
      info_cb: See Download. Usually no messages.

    Returns:
      Response to a download request, normally nothing.
    """
    return self._simple_command('flash', arg=partition, info_cb=info_cb,
                                timeout_ms=timeout_ms)

  def erase(self, partition, timeout_ms=None):
    """Erases the given partition."""
    self._simple_command('erase', arg=partition, timeout_ms=timeout_ms)

  def get_var(self, var, info_cb=DEFAULT_MESSAGE_CALLBACK):
    """Returns the given variable's definition.

    Args:
      var: A variable the bootloader tracks, such as version.
      info_cb: See Download. Usually no messages.
    Returns:
      Value of var according to the current bootloader.
    """
    return self._simple_command('getvar', arg=var, info_cb=info_cb)

  def oem(self, command, timeout_ms=None, info_cb=DEFAULT_MESSAGE_CALLBACK):
    """Executes an OEM command on the device.

    Args:
      command: The command to execute, such as 'poweroff' or 'bootconfig read'.
      timeout_ms: Optional timeout in milliseconds to wait for a response.
      info_cb: See Download. Messages vary based on command.
    Returns:
      The final response from the device.
    """
    return self._simple_command(
        'oem %s' % command, timeout_ms=timeout_ms, info_cb=info_cb)

  def continue_(self):
    """Continues execution past fastboot into the system."""
    return self._simple_command('continue')

  def reboot(self, target_mode=None, timeout_ms=None):
    """Reboots the device.

    Args:
        target_mode: Normal reboot when unspecified (or None). Can specify
            other target modes, such as 'recovery' or 'bootloader'.
        timeout_ms: Optional timeout in milliseconds to wait for a response.
    Returns:
        Usually the empty string. Depends on the bootloader and the target_mode.
    """
    return self._simple_command('reboot', arg=target_mode,
                                timeout_ms=timeout_ms)

  def reboot_bootloader(self, timeout_ms=None):
    """Reboots into the bootloader, usually equiv to Reboot('bootloader')."""
    return self._simple_command('reboot-bootloader', timeout_ms=timeout_ms)



"""Module containing canned measurement validators.

Additional validators may be registered by passing them to the Register()
method.  They can then be accessed directly as attributes on the validators
module, and will typically be a type, instances of which are callable:

  from openhtf.util import validators
  from openhtf.util import measurements

  class MyLessThanValidator(ValidatorBase):
    def __init__(self, limit):
      self.limit = limit

    # This will be invoked to test if the measurement is 'PASS' or 'FAIL'.
    def __call__(self, value):
      return value < self.limit

  # Name defaults to the validator's __name__ attribute unless overridden.
  validators.register(MyLessThanValidator, name='LessThan')

  # Now you can refer to the validator by name directly on measurements.
  @measurements.measures(
      measurements.Measurement('my_measurement').LessThan(4))
  def MyPhase(test):
    test.measurements.my_measurement = 5  # Will have outcome 'FAIL'

If implemented as a class, inherit from a suitable base class defined in this
module; such validators may have specialized handling by the infrastructure that
you can leverage.

For simpler validators, you don't need to register them at all, you can
simply attach them to the Measurement with the .with_validator() method:

  def LessThan4(value):
    return value < 4

  @measurements.measures(
      measurements.Measurement('my_measurement).with_validator(LessThan4))
  def MyPhase(test):
    test.measurements.my_measurement = 5  # Will also 'FAIL'

Notes:

Note the extra level of indirection when registering a validator. This allows
you to parameterize your validator (like in the LessThan example) when it is
being applied to the measurement.  If you don't need this level of indirection,
it's recommended that you simply use .with_validator() instead.

Also note the validator will be str()'d in the output, so if you want a
meaningful description of what it does, you should implement a __str__ method.

Validators must also be deepcopy()'able, and may need to implement __deepcopy__
if they are implemented by a class that has internal state that is not copyable
by the default copy.deepcopy().
"""

import abc
import numbers
import re
from future.utils import with_metaclass
from openhtf import util
import six

_VALIDATORS = {}


def register(validator, name=None):
  name = name or validator.__name__
  if name in _VALIDATORS:
    raise ValueError('Duplicate validator name', name)
  _VALIDATORS[name] = validator
  return validator


def has_validator(name):
  return name in _VALIDATORS


def create_validator(name, *args, **kwargs):
  return _VALIDATORS[name](*args, **kwargs)

_identity = lambda x: x


class ValidatorBase(with_metaclass(abc.ABCMeta, object)):
  @abc.abstractmethod
  def __call__(self, value):
    """Should validate value, returning a boolean result."""


class RangeValidatorBase(with_metaclass(abc.ABCMeta, ValidatorBase)):
  @abc.abstractproperty
  def minimum(self):
    """Should return the minimum, inclusive value of the range."""

  @abc.abstractproperty
  def maximum(self):
    """Should return the maximum, inclusive value of the range."""


# Built-in validators below this line

class InRange(RangeValidatorBase):
  """Validator to verify a numeric value is within a range."""

  def __init__(self, minimum=None, maximum=None, type=None):
    if minimum is None and maximum is None:
      raise ValueError('Must specify minimum, maximum, or both')
    if (minimum is not None and maximum is not None
        and isinstance(minimum, numbers.Number)
        and isinstance(maximum, numbers.Number)
        and minimum > maximum):
      raise ValueError('Minimum cannot be greater than maximum')
    self._minimum = minimum
    self._maximum = maximum
    self._type = type

  @property
  def minimum(self):
    converter = self._type if self._type is not None else _identity
    return converter(self._minimum)

  @property
  def maximum(self):
    converter = self._type if self._type is not None else _identity
    return converter(self._maximum)

  def with_args(self, **kwargs):
    return type(self)(
        minimum=util.format_string(self._minimum, kwargs),
        maximum=util.format_string(self._maximum, kwargs),
        type=self._type,
    )

  def __call__(self, value):
    if value is None:
      return False
    import math
    # Check for nan
    if math.isnan(value):
      return False
    if self.minimum is not None and value < self.minimum:
      return False
    if self.maximum is not None and value > self.maximum:
      return False
    return True

  def __str__(self):
    assert self._minimum is not None or self._maximum is not None
    if self._minimum is not None and self._maximum is not None:
      if self._minimum == self._maximum:
        return 'x == %s' % self._minimum
      return '%s <= x <= %s' % (self._minimum, self._maximum)
    if self._minimum is not None:
      return '%s <= x' % self._minimum
    if self._maximum is not None:
      return 'x <= %s' % self._maximum

  def __eq__(self, other):
    return (isinstance(other, type(self)) and
            self.minimum == other.minimum and self.maximum == other.maximum)

  def __ne__(self, other):
    return not self == other

in_range = InRange  # pylint: disable=invalid-name
register(in_range, name='in_range')


@register
def equals(value, type=None):
  if isinstance(value, numbers.Number):
    return InRange(minimum=value, maximum=value, type=type)
  elif isinstance(value, six.string_types):
    assert type is None or issubclass(type, six.string_types), (
        'Cannot use a non-string type when matching a string')
    return matches_regex('^{}$'.format(re.escape(value)))
  else:
    return Equals(value, type=type)


class Equals(object):
  """Validator to verify an object is equal to the expected value."""

  def __init__(self, expected, type=None):
    self._expected = expected
    self._type = type

  @property
  def expected(self):
    converter = self._type if self._type is not None else _identity
    return converter(self._expected)

  def __call__(self, value):
    return value == self.expected

  def __str__(self):
    return "'x' is equal to '%s'" % self._expected

  def __eq__(self, other):
    return isinstance(other, type(self)) and self.expected == other.expected


class RegexMatcher(object):
  """Validator to verify a string value matches a regex."""

  def __init__(self, regex, compiled_regex):
    self._compiled = compiled_regex
    self.regex = regex

  def __call__(self, value):
    return self._compiled.match(str(value)) is not None

  def __deepcopy__(self, dummy_memo):
    return type(self)(self.regex, self._compiled)

  def __str__(self):
    return "'x' matches /%s/" % self.regex

  def __eq__(self, other):
    return isinstance(other, type(self)) and self.regex == other.regex

  def __ne__(self, other):
    return not self == other


@register
def matches_regex(regex):
  return RegexMatcher(regex, re.compile(regex))


class WithinPercent(RangeValidatorBase):
  """Validates that a number is within percent of a value."""

  def __init__(self, expected, percent):
    if percent < 0:
      raise ValueError('percent argument is {}, must be >0'.format(percent))
    self.expected = expected
    self.percent = percent

  @property
  def _applied_percent(self):
    return abs(self.expected * self.percent / 100.0)

  @property
  def minimum(self):
    return self.expected - self._applied_percent

  @property
  def maximum(self):
    return self.expected + self._applied_percent

  def __call__(self, value):
    return self.minimum <= value <= self.maximum

  def __str__(self):
    return "'x' is within {}% of {}".format(self.percent, self.expected)

  def __eq__(self, other):
    return (isinstance(other, type(self)) and
            self.expected == other.expected and
            self.percent == other.percent)

  def __ne__(self, other):
    return not self == other


@register
def within_percent(expected, percent):
  return WithinPercent(expected, percent)



# Copyright 2018 Google Inc. All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import logging

import sockjs.tornado

from openhtf.util import classproperty

_LOG = logging.getLogger(__name__)


class PubSub(sockjs.tornado.SockJSConnection):
  """Generic pub/sub based on SockJS connections."""

  @classproperty
  def _lock(cls):  # pylint: disable=no-self-argument
    """Ensure subclasses don't share subscriber locks by forcing override."""
    raise AttributeError(
        'The PubSub class should not be instantiated directly. '
        'Instead, subclass it and override the _lock attribute.')

  @classproperty
  def subscribers(cls):  # pylint: disable=no-self-argument
    """Ensure subclasses don't share subscribers by forcing override."""
    raise AttributeError(
        'The PubSub class should not be instantiated directly. '
        'Instead, subclass it and override the subscribers attribute.')

  @classmethod
  def publish(cls, message, client_filter=None):
    """Publish messages to subscribers.

    Args:
      message: The message to publish.
      client_filter: A filter function to call passing in each client. Only
                     clients for whom the function returns True will have the
                     message sent to them.
    """
    with cls._lock:
      for client in cls.subscribers:
        if (not client_filter) or client_filter(client):
          client.send(message)

  def on_open(self, info):
    _LOG.debug('New subscriber from %s.', info.ip)
    with self._lock:
      self.subscribers.add(self)
    self.on_subscribe(info)

  def on_close(self):
    _LOG.debug('A client unsubscribed.')
    with self._lock:
      self.subscribers.remove(self)
    self.on_unsubscribe()

  def on_subscribe(self, info):
    """Called when new clients subscribe. Subclasses can override."""
    pass

  def on_unsubscribe(self):
    """Called when clients unsubscribe. Subclasses can override."""
    pass



"""Utils to convert OpenHTF TestRecord to test_runs_pb2 proto.

MULTIDIM_JSON schema:
{
  "$schema": "http://json-schema.org/draft-04/schema#",
  "title": "Multi-dimensional test parameter",
  "type": "object",
  "properties": {
    "outcome": {"enum": ["PASS", "FAIL", "ERROR"]},
    "name": {"type": "string"},
    "dimensions": {
      "type": array,
      "minItems": 1,
      "items": {
        "type": "object",
        "properties": {
          "uom_code": {"type": "string"},
          "uom_suffix": {"type": "string"}
        }
      }
    },
    "values": {
      "type": "array",
      "items": {}
    }
  }
}
"""

import json
import logging
import numbers

from openhtf.core import measurements
from openhtf.core import test_record
from openhtf.output.callbacks import json_factory
from openhtf.util import validators

from openhtf.output.proto import test_runs_pb2

try:
  from past.types import unicode  # pylint: disable=redefined-builtin,g-import-not-at-top
except ImportError:
  pass

import six  # pylint: disable=g-import-not-at-top,g-bad-import-order


# pylint: disable=no-member
MIMETYPE_MAP = {
    'image/jpeg': test_runs_pb2.JPG,
    'image/png': test_runs_pb2.PNG,
    'audio/x-wav': test_runs_pb2.WAV,
    'text/plain': test_runs_pb2.TEXT_UTF8,
    'image/tiff': test_runs_pb2.TIFF,
    'video/mp4': test_runs_pb2.MP4,
}
OUTCOME_MAP = {
    test_record.Outcome.ERROR: test_runs_pb2.ERROR,
    test_record.Outcome.FAIL: test_runs_pb2.FAIL,
    test_record.Outcome.PASS: test_runs_pb2.PASS,
    test_record.Outcome.TIMEOUT: test_runs_pb2.ERROR,
    test_record.Outcome.ABORTED: test_runs_pb2.ERROR,
}

UOM_CODE_MAP = {
    u.GetOptions().Extensions[
        test_runs_pb2.uom_code]: num
    for num, u in six.iteritems(
        test_runs_pb2.Units.UnitCode.DESCRIPTOR.values_by_number)
}
# pylint: enable=no-member

# Control how many flattened parameters we'll output per multidimensional
# measurement.
MAX_PARAMS_PER_MEASUREMENT = 100


# pylint: disable=invalid-name
def _populate_header(record, testrun):
  """Populate header-like info in testrun from record.

  Mostly obvious, some stuff comes from metadata, see docstring of
  _test_run_from_test_record for details.

  Args:
    record: the OpenHTF TestRecord being converted
    testrun: a TestRun proto
  """
  testrun.dut_serial = record.dut_id
  testrun.tester_name = record.station_id
  if 'test_name' in record.metadata:
    testrun.test_info.name = record.metadata['test_name']
  else:
    # Default to copying tester_name into test_info.name.
    testrun.test_info.name = record.station_id
  if 'test_description' in record.metadata:
    testrun.test_info.description = record.metadata['test_description']
  if 'test_version' in record.metadata:
    testrun.test_info.version_string = record.metadata['test_version']
  testrun.test_status = OUTCOME_MAP[record.outcome]
  testrun.start_time_millis = record.start_time_millis
  testrun.end_time_millis = record.end_time_millis
  if 'run_name' in record.metadata:
    testrun.run_name = record.metadata['run_name']
  for details in record.outcome_details:
    testrun_code = testrun.failure_codes.add()
    testrun_code.code = details.code
    testrun_code.details = details.description
  for phase in record.phases:
    testrun_phase = testrun.phases.add()
    testrun_phase.name = phase.name
    testrun_phase.description = phase.codeinfo.sourcecode
    testrun_phase.timing.start_time_millis = phase.start_time_millis
    testrun_phase.timing.end_time_millis = phase.end_time_millis
  if 'config' in record.metadata:
    attachment = testrun.info_parameters.add()
    attachment.name = 'config'
    attachment.value_binary = json.dumps(
        record.metadata['config'], sort_keys=True, indent=4).encode('utf-8')


def _ensure_unique_parameter_name(name, used_parameter_names):
  while name in used_parameter_names:
    name += '_'  # Hack to avoid collisions between phases.
  used_parameter_names.add(name)
  return name


def _attach_json(record, testrun):
  """Attach a copy of the JSON-ified record as an info parameter.

  Save a copy of the JSON-ified record in an attachment so we can access
  un-mangled fields later if we want.  Remove attachments since those get
  copied over and can potentially be quite large.

  Args:
    record: the OpenHTF TestRecord being converted
    testrun: a TestRun proto
  """
  record_json = json_factory.OutputToJSON(
      inline_attachments=False,
      sort_keys=True, indent=2).serialize_test_record(record)
  testrun_param = testrun.info_parameters.add()
  testrun_param.name = 'OpenHTF_record.json'
  testrun_param.value_binary = record_json.encode('utf-8')
  # pylint: disable=no-member
  testrun_param.type = test_runs_pb2.TEXT_UTF8
  # pylint: enable=no-member


def _extract_attachments(phase, testrun, used_parameter_names):
  """Extract attachments, just copy them over."""
  for name, attachment in sorted(six.iteritems(phase.attachments)):
    attachment_data, mimetype = attachment
    name = _ensure_unique_parameter_name(name, used_parameter_names)
    testrun_param = testrun.info_parameters.add()
    testrun_param.name = name
    if isinstance(attachment_data, unicode):
      attachment_data = attachment_data.encode('utf-8')
    testrun_param.value_binary = attachment_data
    if mimetype in MIMETYPE_MAP:
      testrun_param.type = MIMETYPE_MAP[mimetype]
    else:
      # pylint: disable=no-member
      testrun_param.type = test_runs_pb2.BINARY
      # pylint: enable=no-member


def _mangle_measurement(name, measured_value, measurement, mangled_parameters,
                        attachment_name):  # pylint: disable=g-doc-args
  """Flatten parameters for backwards compatibility, watch for collisions.

  We generate these by doing some name mangling, using some sane limits for
  very large multidimensional measurements.
  """
  for coord, val in list(measured_value.value_dict.items(
      ))[:MAX_PARAMS_PER_MEASUREMENT]:
    # Mangle names so they look like 'myparameter_Xsec_Ynm_ZHz'
    mangled_name = '_'.join([name] + [
        '%s%s' % (
            dim_val,
            dim_units.suffix if dim_units.suffix else '') for
        dim_val, dim_units in zip(
            coord, measurement.dimensions)])
    while mangled_name in mangled_parameters:
      logging.warning('Mangled name %s already in use', mangled_name)
      mangled_name += '_'
    mangled_param = test_runs_pb2.TestParameter()
    mangled_param.name = mangled_name
    mangled_param.associated_attachment = attachment_name
    mangled_param.description = (
        'Mangled parameter from measurement %s with dimensions %s' % (
            name, tuple(d.suffix for d in measurement.dimensions)))

    if isinstance(val, numbers.Number):
      mangled_param.numeric_value = float(val)
    else:
      mangled_param.text_value = str(val)
    # Check for validators we know how to translate.
    for validator in measurement.validators:
      mangled_param.description += '\nValidator: ' + str(validator)

    if measurement.units and measurement.units.code in UOM_CODE_MAP:
      mangled_param.unit_code = UOM_CODE_MAP[measurement.units.code]
    mangled_parameters[mangled_name] = mangled_param


def _extract_parameters(record, testrun, used_parameter_names):
  """Extract parameters from phases."""
  # Generate mangled parameters afterwards so we give real measurements priority
  # getting names.
  mangled_parameters = {}
  for phase in record.phases:
    _extract_attachments(phase, testrun, used_parameter_names)
    for name, measurement in sorted(six.iteritems(phase.measurements)):
      tr_name = _ensure_unique_parameter_name(name, used_parameter_names)
      testrun_param = testrun.test_parameters.add()
      testrun_param.name = tr_name
      if measurement.docstring:
        testrun_param.description = measurement.docstring
      if measurement.units and measurement.units.code in UOM_CODE_MAP:
        testrun_param.unit_code = UOM_CODE_MAP[measurement.units.code]

      if measurement.outcome == measurements.Outcome.PASS:
        testrun_param.status = test_runs_pb2.PASS
      elif (not measurement.measured_value
            or not measurement.measured_value.is_value_set):
        testrun_param.status = test_runs_pb2.ERROR
        continue
      else:
        testrun_param.status = test_runs_pb2.FAIL

      value = None
      if measurement.measured_value.is_value_set:
        value = measurement.measured_value.value
      else:
        testrun_param.status = test_runs_pb2.ERROR
      if measurement.dimensions is None:
        # Just a plain ol' value.
        if isinstance(value, numbers.Number):
          testrun_param.numeric_value = float(value)
        else:
          testrun_param.text_value = str(value)
        # Check for validators we know how to translate.
        for validator in measurement.validators:
          if isinstance(validator, validators.RangeValidatorBase):
            if validator.minimum is not None:
              testrun_param.numeric_minimum = float(validator.minimum)
            if validator.maximum is not None:
              testrun_param.numeric_maximum = float(validator.maximum)
          elif isinstance(validator, validators.RegexMatcher):
            testrun_param.expected_text = validator.regex
          else:
            testrun_param.description += '\nValidator: ' + str(validator)
      else:
        attachment = testrun.info_parameters.add()
        attachment.name = 'multidim_%s' % name
        dims = [{
            'uom_suffix': d.suffix,
            'uom_code': d.code}
                for d in measurement.dimensions]
        # Refer to the module docstring for the expected schema.
        attachment.value_binary = json.dumps({
            'outcome': str(testrun_param.status), 'name': name,
            'dimensions': dims,
            'value': value
        }, sort_keys=True).encode('utf-8')
        attachment.type = test_runs_pb2.MULTIDIM_JSON
        _mangle_measurement(
            name, measurement.measured_value, measurement, mangled_parameters,
            attachment.name)
      if testrun_param.status == test_runs_pb2.FAIL:
        testrun_code = testrun.failure_codes.add()
        testrun_code.code = testrun_param.name
        if measurement.dimensions is None:
          if isinstance(testrun_param.numeric_value, float):
            testrun_code.details = str(testrun_param.numeric_value)
          else:
            testrun_code.details = testrun_param.text_value
  return mangled_parameters


def _add_mangled_parameters(testrun, mangled_parameters, used_parameter_names):
  """Add any mangled parameters we generated from multidim measurements."""
  for mangled_name, mangled_param in sorted(six.iteritems(mangled_parameters)):
    if mangled_name != _ensure_unique_parameter_name(mangled_name,
                                                     used_parameter_names):
      logging.warning('Mangled name %s in use by non-mangled parameter',
                      mangled_name)
    testrun_param = testrun.test_parameters.add()
    testrun_param.CopyFrom(mangled_param)


def _add_log_lines(record, testrun):
  """Copy log records over, this is a fairly straightforward mapping."""
  for log in record.log_records:
    testrun_log = testrun.test_logs.add()
    testrun_log.timestamp_millis = log.timestamp_millis
    testrun_log.log_message = log.message
    testrun_log.logger_name = log.logger_name
    testrun_log.levelno = log.level
    # pylint: disable=no-member
    if log.level <= logging.DEBUG:
      testrun_log.level = test_runs_pb2.TestRunLogMessage.DEBUG
    elif log.level <= logging.INFO:
      testrun_log.level = test_runs_pb2.TestRunLogMessage.INFO
    elif log.level <= logging.WARNING:
      testrun_log.level = test_runs_pb2.TestRunLogMessage.WARNING
    elif log.level <= logging.ERROR:
      testrun_log.level = test_runs_pb2.TestRunLogMessage.ERROR
    elif log.level <= logging.CRITICAL:
      testrun_log.level = test_runs_pb2.TestRunLogMessage.CRITICAL
    # pylint: enable=no-member
    testrun_log.log_source = log.source
    testrun_log.lineno = log.lineno


def test_run_from_test_record(record):
  """Create a TestRun proto from an OpenHTF TestRecord.

  Most fields are just copied over, some are pulled out of metadata (listed
  below), and measurements are munged a bit for backwards compatibility.

  Metadata fields:
    'test_description': TestInfo's description field.
    'test_version': TestInfo's version_string field.
    'test_name': TestInfo's name field.
    'run_name': TestRun's run_name field.
    'operator_name': TestRun's operator_name field.

  Args:
    record: the OpenHTF TestRecord being converted

  Returns:
    An instance of the TestRun proto for the given record.
  """
  testrun = test_runs_pb2.TestRun()
  _populate_header(record, testrun)
  _attach_json(record, testrun)

  used_parameter_names = set(['OpenHTF_record.json'])
  mangled_parameters = _extract_parameters(record, testrun,
                                           used_parameter_names)
  _add_mangled_parameters(testrun, mangled_parameters, used_parameter_names)
  _add_log_lines(record, testrun)
  return testrun



# Copyright 2018 Google Inc. All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Modified by Tack Verification, 2020

"""Phase Groups in OpenHTF.

Phase Groups are collections of Phases that are used to control phase
shortcutting due to terminal errors to better guarentee when teardown phases
run.

PhaseGroup instances have three primary member fields:
  `setup`: a list of phases, run first.  If these phases are all non-terminal,
      the PhaseGroup is entered.
  `main`: a list of phases, run after the setup phases as long as those are
      non-terminal.  If any of these phases are terminal, then the rest of the
      main phases will be skipped.
  `teardown`: a list of phases, guarenteed to run after the main phases as long
      as the PhaseGroup was entered.  If any are terminal, other teardown phases
      will continue to be run.  One exception is that a second CTRL-C sent to
      the main thread will abort all teardown phases.
There is one optional field:
  `name`: str, an arbitrary description used for logging.

PhaseGroup instances can be nested inside of each other.  A PhaseGroup is
terminal if any of its Phases or further nested PhaseGroups are also terminal.
"""

import collections
import functools

import mutablerecords

from openhtf.core import phase_descriptor
from openhtf.core import test_record


class PhaseGroup(mutablerecords.Record(
    'PhaseGroup', [], {
        'setup': tuple,
        'main': tuple,
        'teardown': tuple,
        'name': None,
    })):
  """Phase group with guaranteed end phase running.

  If the setup phases all continue, then the main phases and teardown phases are
  run. Even if any main phase or teardown phases has a terminal error, all the
  teardown phases are guaranteed to be run.
  """

  def __init__(self, setup=None, main=None, teardown=None, name=None):
    if not setup:
      setup = ()
    elif isinstance(setup, PhaseGroup):
      setup = (setup,)
    if not main:
      main = ()
    elif isinstance(main, PhaseGroup):
      main = (main,)
    if not teardown:
      teardown = ()
    elif isinstance(teardown, PhaseGroup):
      teardown = (teardown,)
      
    setup = tuple(self.fn_as_type(p, phase_descriptor.PhaseType.SETUP) for p in setup)
    main = tuple(self.fn_as_type(p, phase_descriptor.PhaseType.MAIN) for p in main)
    teardown = tuple(self.fn_as_type(p, phase_descriptor.PhaseType.TEARDOWN) for p in teardown)
      
    super(PhaseGroup, self).__init__(
        setup=tuple(setup), main=tuple(main), teardown=tuple(teardown),
        name=name)

  def fn_as_type(self, fn, phase_type):
    try:
      return fn.as_type(phase_type)
    except AttributeError:
      return fn

  @classmethod
  def convert_if_not(cls, phases_or_groups):
    """Convert list of phases or groups into a new PhaseGroup if not already."""
    if isinstance(phases_or_groups, PhaseGroup):
      return mutablerecords.CopyRecord(phases_or_groups)

    flattened = flatten_phases_and_groups(phases_or_groups)
    return cls(main=flattened)

  @classmethod
  def with_context(cls, setup_phases, teardown_phases):
    """Create PhaseGroup creator function with setup and teardown phases.

    Args:
      setup_phases: list of phase_descriptor.PhaseDescriptors/PhaseGroups/
          callables/iterables, phases to run during the setup for the PhaseGroup
          returned from the created function.
      teardown_phases: list of phase_descriptor.PhaseDescriptors/PhaseGroups/
          callables/iterables, phases to run during the teardown for the
          PhaseGroup returned from the created function.

    Returns:
      Function that takes *phases and returns a PhaseGroup with the predefined
      setup and teardown phases, with *phases as the main phases.
    """
    setup = flatten_phases_and_groups(setup_phases)
    teardown = flatten_phases_and_groups(teardown_phases)

    def _context_wrapper(*phases):
      return cls(setup=setup,
                 main=flatten_phases_and_groups(phases),
                 teardown=teardown)
    return _context_wrapper

  @classmethod
  def with_setup(cls, *setup_phases):
    """Create PhaseGroup creator function with predefined setup phases."""
    return cls.with_context(setup_phases, [])

  @classmethod
  def with_teardown(cls, *teardown_phases):
    """Create PhaseGroup creator function with predefined teardown phases."""
    return cls.with_context([], teardown_phases)

  def combine(self, other, name=None):
    """Combine with another PhaseGroup and return the result."""
    return PhaseGroup(
        setup=self.setup + other.setup,
        main=self.main + other.main,
        teardown=self.teardown + other.teardown,
        name=name)

  def wrap(self, main_phases, name=None):
    """Returns PhaseGroup with additional main phases."""
    new_main = list(self.main)
    if isinstance(main_phases, collections.abc.Iterable):
      new_main.extend(main_phases)
    else:
      new_main.append(main_phases)
    return PhaseGroup(
        setup=self.setup,
        main=new_main,
        teardown=self.teardown,
        name=name)

  def transform(self, transform_fn):
    return PhaseGroup(
        setup=[transform_fn(p) for p in self.setup],
        main=[transform_fn(p) for p in self.main],
        teardown=[transform_fn(p) for p in self.teardown],
        name=self.name)

  def with_args(self, **kwargs):
    """Send known keyword-arguments to each contained phase the when called."""
    return self.transform(functools.partial(optionally_with_args, **kwargs))

  def with_plugs(self, **subplugs):
    """Substitute only known plugs for placeholders for each contained phase."""
    return self.transform(functools.partial(optionally_with_plugs, **subplugs))

  def as_type(self, type):
    """Same interface as PhaseDescriptors"""
    return self
  
  def as_depth(self, depth):
    """Same interface as PhaseDescriptors"""
    return self

  def _iterate(self, phases):
    for phase in phases:
      if isinstance(phase, PhaseGroup):
        for p in phase:
          yield p
      else:
        yield phase

  def __iter__(self):
    """Iterate directly over the phases."""
    for phase in self._iterate(self.setup):
      yield phase
    for phase in self._iterate(self.main):
      yield phase
    for phase in self._iterate(self.teardown):
      yield phase

  def flatten(self):
    """Internally flatten out nested iterables."""
    return PhaseGroup(
        setup=flatten_phases_and_groups(self.setup),
        main=flatten_phases_and_groups(self.main),
        teardown=flatten_phases_and_groups(self.teardown),
        name=self.name)

  def build_tree(self):
    return build_phase_group_tree(self)

  def load_code_info(self, with_source=False):
    """Load coded info for all contained phases."""
    return PhaseGroup(
        setup=load_code_info(self.setup, with_source=with_source),
        main=load_code_info(self.main, with_source=with_source),
        teardown=load_code_info(self.teardown, with_source=with_source),
        name=self.name)


def load_code_info(phases_or_groups, with_source=False):
  """Recursively load code info for a PhaseGroup or list of phases or groups."""
  if isinstance(phases_or_groups, PhaseGroup):
    return phases_or_groups.load_code_info(with_source=with_source)
  ret = []
  for phase in phases_or_groups:
    if isinstance(phase, PhaseGroup):
      ret.append(phase.load_code_info(with_source=with_source))
    else:
      ret.append(
          mutablerecords.CopyRecord(
              phase, code_info=test_record.CodeInfo.for_function(phase.func, with_source=with_source)))
  return ret


def flatten_phases_and_groups(phases_or_groups):
  """Recursively flatten nested lists for the list of phases or groups."""
  if isinstance(phases_or_groups, PhaseGroup):
    phases_or_groups = [phases_or_groups]
  ret = []
  
  _flatten_functions(phases_or_groups, 
    on_phase_group = lambda phase: ret.append(phase.flatten()),
    on_iterable = lambda phase: ret.extend(flatten_phases_and_groups(phase)),
    on_else = lambda phase: ret.append(phase_descriptor.PhaseDescriptor.wrap_or_copy(phase))
  )
  
  return ret

def build_phase_group_tree(phases_or_groups, _depth=0):
  if isinstance(phases_or_groups, PhaseGroup):
    phases_or_groups = [phases_or_groups]
    
  tree = []
  
  _flatten_functions(phases_or_groups, 
    on_phase_group = lambda phase: tree.append(build_phase_group_tree(phase.setup + phase.main + phase.teardown, _depth=_depth+1)),
    on_iterable = lambda phase: tree.append(build_phase_group_tree(phase, _depth=_depth+1)),
    on_else = lambda phase: tree.append(phase_descriptor.PhaseDescriptor.wrap_or_copy(phase).as_depth(_depth))
  )
  
  return tree
  

def _flatten_functions(list_to_flatten, on_phase_group, on_iterable, on_else):
  for item in list_to_flatten:
    if isinstance(item, PhaseGroup):
      on_phase_group(item)
    elif isinstance(item, collections.abc.Iterable):
      on_iterable(item)
    else:
      on_else(item)
      
  


def optionally_with_args(phase, **kwargs):
  """Apply only the args that the phase knows.

  If the phase has a **kwargs-style argument, it counts as knowing all args.

  Args:
    phase: phase_descriptor.PhaseDescriptor or PhaseGroup or callable, or
        iterable of those, the phase or phase group (or iterable) to apply
        with_args to.
    **kwargs: arguments to apply to the phase.

  Returns:
    phase_descriptor.PhaseDescriptor or PhaseGroup or iterable with the updated
    args.
  """
  if isinstance(phase, PhaseGroup):
    return phase.with_args(**kwargs)
  if isinstance(phase, collections.abc.Iterable):
    return [optionally_with_args(p, **kwargs) for p in phase]

  if not isinstance(phase, phase_descriptor.PhaseDescriptor):
    phase = phase_descriptor.PhaseDescriptor.wrap_or_copy(phase)
  return phase.with_known_args(**kwargs)


def optionally_with_plugs(phase, **subplugs):
  """Apply only the with_plugs that the phase knows.

  This will determine the subset of plug overrides for only plugs the phase
  actually has.

  Args:
    phase: phase_descriptor.PhaseDescriptor or PhaseGroup or callable, or
        iterable of those, the phase or phase group (or iterable) to apply the
        plug changes to.
    **subplugs: mapping from plug name to derived plug class, the subplugs to
        apply.

  Raises:
    openhtf.plugs.InvalidPlugError: if a specified subplug class is not a valid
        replacement for the specified plug name.

  Returns:
    phase_descriptor.PhaseDescriptor or PhaseGroup or iterable with the updated
    plugs.
  """
  if isinstance(phase, PhaseGroup):
    return phase.with_plugs(**subplugs)
  if isinstance(phase, collections.abc.Iterable):
    return [optionally_with_plugs(p, **subplugs) for p in phase]

  if not isinstance(phase, phase_descriptor.PhaseDescriptor):
    phase = phase_descriptor.PhaseDescriptor.wrap_or_copy(phase)
  return phase.with_known_plugs(**subplugs)



"""Module for outputting test record to JSON-formatted files."""

import base64
import json

from openhtf.output import callbacks
from openhtf.util import data
import six


class OutputToJSON(callbacks.OutputToFile):
  """Return an output callback that writes JSON Test Records.
  Example filename_patterns might be:
    '/data/test_records/{dut_id}.{metadata[test_name]}.json', indent=4)) or
    '/data/test_records/%(dut_id)s.%(start_time_millis)s'
  To use this output mechanism:
    test = openhtf.Test(PhaseOne, PhaseTwo)
    test.add_output_callback(openhtf.output.callbacks.OutputToJson(
        '/data/test_records/{dut_id}.{metadata[test_name]}.json'))

  Args:
    filename_pattern: A format string specifying the filename to write to,
      will be formatted with the Test Record as a dictionary.  May also be a
      file-like object to write to directly.
    inline_attachments: Whether attachments should be included inline in the
      output. Set to False if you expect to have large binary attachments. If
      True (the default), then attachments are base64 encoded to allow for
      binary data that's not supported by JSON directly.
  """

  def __init__(self, filename_pattern=None, inline_attachments=True, **kwargs):
    super(OutputToJSON, self).__init__(filename_pattern)
    self.inline_attachments = inline_attachments

    # Conform strictly to the JSON spec by default.
    kwargs.setdefault('allow_nan', False)
    self.allow_nan = kwargs['allow_nan']
    self.json_encoder = json.JSONEncoder(**kwargs)

  def serialize_test_record(self, test_record):
    return self.json_encoder.encode(self.convert_to_dict(test_record))

  def convert_to_dict(self, test_record):
    as_dict = data.convert_to_base_types(test_record,
                                         json_safe=(not self.allow_nan))
    if self.inline_attachments:
      for phase, original_phase in zip(as_dict['phases'], test_record.phases):
        for name, attachment in six.iteritems(phase['attachments']):
          original_data = original_phase.attachments[name].data
          attachment['data'] = base64.standard_b64encode(original_data).decode('utf-8')
    return as_dict
